@@ -44,6 +44,7 @@ const Mainloop = imports.mainloop;
 const MimeHandler = imports.ui.mimeHandler;
 const Constants = imports.util.constants;
 const SpinnerBox = imports.ui.spinnerBox;
+const Utils = imports.ui.utils;
 
 const Sushi = imports.gi.Sushi;
 
@@ -75,9 +76,12 @@ MainWindow.prototype = {
     },
 
     _createGtkWindow : function() {
+        this._settings = new Gio.Settings({ schema_id: 'org.gnome.sushi' });
+        this._clientDecorated = this._settings.get_boolean('client-decoration');
+
         this._gtkWindow = new Gtk.Window({ type: Gtk.WindowType.TOPLEVEL,
                                            focusOnMap: true,
-                                           decorated: false,
+                                           decorated: !this._clientDecorated,
                                            hasResizeGrip: false,
                                            skipPagerHint: true,
                                            skipTaskbarHint: true,
@@ -87,8 +91,15 @@ MainWindow.prototype = {
         let screen = Gdk.Screen.get_default();
         this._gtkWindow.set_visual(screen.get_rgba_visual());
 
-        this._gtkWindow.connect("delete-event",
+        this._gtkWindow.connect('delete-event',
                                 Lang.bind(this, this._onWindowDeleteEvent));
+        this._gtkWindow.connect('realize', Lang.bind(this,
+            function() {
+                // don't support maximize and minimize
+                this._gtkWindow.get_window().set_functions(Gdk.WMFunction.MOVE |
+                                                           Gdk.WMFunction.RESIZE |
+                                                           Gdk.WMFunction.CLOSE);
+            }));
     },
 
     _createClutterEmbed : function() {
@@ -105,17 +116,19 @@ MainWindow.prototype = {
                                                   green: 0,
                                                   blue: 0,
                                                   alpha: 255 }));
-        this._mainGroup =  new Clutter.Group();
+
+        this._mainLayout = new Clutter.BinLayout();
+        this._mainGroup = new Clutter.Actor({ layout_manager: this._mainLayout });
         this._stage.add_actor(this._mainGroup);
         this._mainGroup.set_opacity(0);
     },
 
     _connectStageSignals : function() {
-        this._stage.connect("key-press-event",
+        this._stage.connect('key-press-event',
                             Lang.bind(this, this._onStageKeyPressEvent));
-        this._stage.connect("button-press-event",
+        this._stage.connect('button-press-event',
                             Lang.bind(this, this._onButtonPressEvent));
-        this._stage.connect("motion-event",
+        this._stage.connect('motion-event',
                             Lang.bind(this, this._onMotionEvent));
     },
 
@@ -136,7 +149,7 @@ MainWindow.prototype = {
             new Clutter.BindConstraint({ source: this._stage,
                                          coordinate: Clutter.BindCoordinate.SIZE }));
 
-        this._mainGroup.add_actor(this._background);
+        this._mainGroup.add_child(this._background);
         this._background.lower_bottom();
     },
 
@@ -144,7 +157,16 @@ MainWindow.prototype = {
         if (this._background)
             return;
 
-        this._background = Sushi.create_rounded_background();
+        if (this._clientDecorated) {
+            this._background = Sushi.create_rounded_background();
+        } else {
+            this._background = new Clutter.Rectangle();
+            this._background.set_color(new Clutter.Color({ red: 0,
+                                                           green: 0,
+                                                           blue: 0,
+                                                           alpha: 255 }));
+        }
+
         this._background.set_opacity(Constants.VIEW_BACKGROUND_OPACITY);
         this._background.add_constraint(
             new Clutter.BindConstraint({ source: this._stage,
@@ -153,7 +175,7 @@ MainWindow.prototype = {
             new Clutter.BindConstraint({ source: this._stage,
                                          coordinate: Clutter.BindCoordinate.SIZE }));
 
-        this._mainGroup.add_actor(this._background);
+        this._mainGroup.add_child(this._background);
         this._background.lower_bottom();
     },
 
@@ -191,7 +213,7 @@ MainWindow.prototype = {
             return false;
         }
 
-        let root_coords = 
+        let root_coords =
             this._gtkWindow.get_window().get_root_coords(win_coords[0],
                                                          win_coords[1]);
 
@@ -216,9 +238,10 @@ MainWindow.prototype = {
     _getTextureSize : function() {
         let screenSize = [ this._gtkWindow.get_window().get_width(),
                            this._gtkWindow.get_window().get_height() ];
+        let yPadding = this._clientDecorated ? Constants.VIEW_PADDING_Y : 0;
 
         let availableWidth = this._isFullScreen ? screenSize[0] : Constants.VIEW_MAX_W - 2 * Constants.VIEW_PADDING_X;
-        let availableHeight = this._isFullScreen ? screenSize[1] : Constants.VIEW_MAX_H - Constants.VIEW_PADDING_Y;
+        let availableHeight = this._isFullScreen ? screenSize[1] : Constants.VIEW_MAX_H - yPadding;
 
         let textureSize = this._renderer.getSizeForAllocation([availableWidth, availableHeight], this._isFullScreen);
 
@@ -228,13 +251,14 @@ MainWindow.prototype = {
     _getWindowSize : function() {
         let textureSize = this._getTextureSize();
         let windowSize = textureSize;
+        let yPadding = this._clientDecorated ? Constants.VIEW_PADDING_Y : 0;
 
         if (textureSize[0] < (Constants.VIEW_MIN - 2 * Constants.VIEW_PADDING_X) &&
-            textureSize[1] < (Constants.VIEW_MIN - Constants.VIEW_PADDING_Y)) {
+            textureSize[1] < (Constants.VIEW_MIN - yPadding)) {
             windowSize = [ Constants.VIEW_MIN, Constants.VIEW_MIN ];
         } else if (!this._isFullScreen) {
             windowSize = [ windowSize[0] + 2 * Constants.VIEW_PADDING_X,
-                           windowSize[1] + Constants.VIEW_PADDING_Y ];
+                           windowSize[1] + yPadding ];
         }
 
         return windowSize;
@@ -252,7 +276,7 @@ MainWindow.prototype = {
         }
 
         if (yFactor == 0) {
-            if (this._isFullScreen && 
+            if (this._isFullScreen &&
                (textureSize[0] > textureSize[1]))
                 yFactor = 0.52;
             else
@@ -289,7 +313,7 @@ MainWindow.prototype = {
         this._renderer.startTimeout();
 
         file.query_info_async
-        (Gio.FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME + "," +
+        (Gio.FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME + ',' +
          Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
          Gio.FileQueryInfoFlags.NONE,
          GLib.PRIORITY_DEFAULT, null,
@@ -327,19 +351,19 @@ MainWindow.prototype = {
 
         this._texture = this._renderer.render();
 
-        this._textureXAlign = 
+        this._textureXAlign =
             new Clutter.AlignConstraint({ source: this._stage,
                                           factor: 0.5 });
         this._textureYAlign =
             new Clutter.AlignConstraint({ source: this._stage,
                                           factor: 0.5,
-                                          "align-axis": Clutter.AlignAxis.Y_AXIS })
+                                          align_axis: Clutter.AlignAxis.Y_AXIS });
 
         this._texture.add_constraint(this._textureXAlign);
         this._texture.add_constraint(this._textureYAlign);
 
         this.refreshSize();
-        this._mainGroup.add_actor(this._texture);
+        this._mainGroup.add_child(this._texture);
     },
 
     /**************************************************************************
@@ -363,12 +387,12 @@ MainWindow.prototype = {
         Tweener.addTween(this._mainGroup,
                          { opacity: 255,
                            time: 0.15,
-                           transition: 'easeOutQuad',
+                           transition: 'easeOutQuad'
                          });
         Tweener.addTween(this._titleGroup,
                          { opacity: 255,
                            time: 0.15,
-                           transition: 'easeOutQuad',
+                           transition: 'easeOutQuad'
                          });
     },
 
@@ -380,11 +404,11 @@ MainWindow.prototype = {
             this._removeToolbarTimeout();
         }
 
-        /* wait for the next stage allocation to fade in the texture 
+        /* wait for the next stage allocation to fade in the texture
          * and background again.
          */
         this._unFullScreenId =
-            this._stage.connect("notify::allocation",
+            this._stage.connect('notify::allocation',
                                 Lang.bind(this, this._onStageUnFullScreen));
 
         /* quickly fade out everything,
@@ -451,11 +475,11 @@ MainWindow.prototype = {
             this._removeToolbarTimeout();
         }
 
-        /* wait for the next stage allocation to fade in the texture 
+        /* wait for the next stage allocation to fade in the texture
          * and background again.
          */
         this._fullScreenId =
-            this._stage.connect("notify::allocation",
+            this._stage.connect('notify::allocation',
                                 Lang.bind(this, this._onStageFullScreen));
 
         /* quickly fade out everything,
@@ -493,24 +517,17 @@ MainWindow.prototype = {
         if (!this._toolbarActor)
             return;
 
+        Utils.alphaGtkWidget(this._toolbarActor.get_widget());
+
         this._toolbarActor.set_reactive(true);
         this._toolbarActor.set_opacity(0);
-        this._mainGroup.add_actor(this._toolbarActor);
-
         this._toolbarActor.add_constraint(
             new Clutter.AlignConstraint({ source: this._stage,
                                           factor: 0.5 }));
 
-        let yConstraint = 
-            new Clutter.BindConstraint({ source: this._stage,
-                                         coordinate: Clutter.BindCoordinate.Y,
-                                         offset: this._stage.height - Constants.TOOLBAR_SPACING });
-        this._toolbarActor.add_constraint(yConstraint);
-
-        this._stage.connect("notify::height",
-                            Lang.bind(this, function() {
-                                yConstraint.set_offset(this._stage.height - Constants.TOOLBAR_SPACING);
-                            }));
+        this._toolbarActor.margin_bottom = Constants.TOOLBAR_SPACING;
+        this._mainLayout.add(this._toolbarActor,
+                             Clutter.BinAlignment.FIXED, Clutter.BinAlignment.END);
     },
 
     _removeToolbarTimeout: function() {
@@ -530,7 +547,7 @@ MainWindow.prototype = {
             Tweener.addTween(this._toolbarActor,
                              { opacity: 200,
                                time: 0.1,
-                               transition: 'easeOutQuad',
+                               transition: 'easeOutQuad'
                              });
         }
 
@@ -569,21 +586,25 @@ MainWindow.prototype = {
             new Clutter.BindConstraint({ source: this._stage,
                                          coordinate: Clutter.BindCoordinate.WIDTH }));
 
-        this._titleLabel = new Gtk.Label({ label: "",
+        // if we don't draw client decorations, just create an empty actor
+        if (!this._clientDecorated)
+            return;
+
+        this._titleLabel = new Gtk.Label({ label: '',
 					   ellipsize: Pango.EllipsizeMode.END,
                                            margin: 6 });
-        this._titleLabel.get_style_context().add_class("np-decoration");
-        
+        this._titleLabel.get_style_context().add_class('np-decoration');
+
         this._titleLabel.show();
         this._titleActor = new GtkClutter.Actor({ contents: this._titleLabel });
 
-        this._quitButton = 
-            new Gtk.Button({ image: new Gtk.Image ({ "icon-size": Gtk.IconSize.MENU,
-                                                     "icon-name": "window-close-symbolic" })});
-        this._quitButton.get_style_context().add_class("np-decoration");
+        this._quitButton =
+            new Gtk.Button({ image: new Gtk.Image ({ icon_size: Gtk.IconSize.MENU,
+                                                     icon_name: 'window-close-symbolic' })});
+        this._quitButton.get_style_context().add_class('np-decoration');
         this._quitButton.show();
 
-        this._quitButton.connect("clicked",
+        this._quitButton.connect('clicked',
                                  Lang.bind(this,
                                            this._clearAndQuit));
 
@@ -627,7 +648,7 @@ MainWindow.prototype = {
         Tweener.addTween(this._titleGroup,
                          { opacity: 0,
                            time: 0.15,
-                           transition: 'easeOutQuad',
+                           transition: 'easeOutQuad'
                          });
 
         Tweener.addTween(this._mainGroup,
@@ -656,6 +677,8 @@ MainWindow.prototype = {
         this._gtkWindow.realize();
         this._gtkWindow.get_window().set_transient_for(this._parent);
         this._gtkWindow.show_all();
+
+        this._gtkWindow.get_window().move_to_current_desktop();
     },
 
     setFile : function(file) {
@@ -670,7 +693,10 @@ MainWindow.prototype = {
     },
 
     setTitle : function(label) {
-        this._titleLabel.set_label(label);
+        if (this._clientDecorated)
+            this._titleLabel.set_label(label);
+
+        this._gtkWindow.set_title(label);
     },
 
     refreshSize : function() {
